Package: netsubsamp
Type: Package
Title: Multivariate Inference of Network Moments by Subsampling
Version: 1.0.0
Authors@R: c(
    person("Mingyu",    "Qi",   role = "aut"),
    person("Chen-Wei",  "Hua",  role = "aut"),
    person("Tianxi",    "Li",   email = "tianxili@umn.edu", role = c("aut", "cre")),
    person("Wen",       "Zhou", role = "aut"))
Description: Implements node subsampling methods for multivariate inference on
    network moments (rescaled motif counts), including: uniform node subsampling
    to approximate the joint distribution of multiple network moments (Algorithm 1);
    externally sparsified moments for density-matched comparisons (Algorithm 2);
    and a two-sample test for unmatchable networks with unequal edge densities
    via a split-and-sparsify subsampling procedure (Algorithm 3). Built-in support
    for V-shape (2-star), triangle, and 3-star motifs, with a user-extensible
    interface for arbitrary additional motifs. Parallel execution is supported
    via 'doParallel' and 'foreach'. Based on Qi, Hua, Li and Zhou (2024)
    <doi:10.48550/arXiv.2409.01599>.
License: GPL-3
Encoding: UTF-8
Imports: foreach, doParallel, parallel, stats
Suggests: Matrix, randnet, testthat (>= 3.0.0)
RoxygenNote: 7.3.1
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-15 14:32:26 UTC; tianxili
Author: Mingyu Qi [aut],
  Chen-Wei Hua [aut],
  Tianxi Li [aut, cre],
  Wen Zhou [aut]
Maintainer: Tianxi Li <tianxili@umn.edu>
Repository: CRAN
Date/Publication: 2026-04-21 18:22:07 UTC
Built: R 4.6.0; ; 2026-04-27 18:47:51 UTC; unix

Package: netsubsamp
Type: Package
Title: Multivariate Inference of Network Moments by Subsampling
Version: 1.0.0
Authors@R: c(
    person("Mingyu",    "Qi",   role = "aut"),
    person("Chen-Wei",  "Hua",  role = "aut"),
    person("Tianxi",    "Li",   email = "tianxili@umn.edu", role = c("aut", "cre")),
    person("Wen",       "Zhou", role = "aut"))
Description: Implements node subsampling methods for multivariate inference on
    network moments (rescaled motif counts), including: uniform node subsampling
    to approximate the joint distribution of multiple network moments (Algorithm 1);
    externally sparsified moments for density-matched comparisons (Algorithm 2);
    and a two-sample test for unmatchable networks with unequal edge densities
    via a split-and-sparsify subsampling procedure (Algorithm 3). Built-in support
    for V-shape (2-star), triangle, and 3-star motifs, with a user-extensible
    interface for arbitrary additional motifs. Parallel execution is supported
    via 'doParallel' and 'foreach'. Based on Qi, Hua, Li and Zhou (2024)
    <doi:10.48550/arXiv.2409.01599>.
License: GPL-3
Encoding: UTF-8
Imports: foreach, doParallel, parallel, stats
Suggests: Matrix, randnet, testthat (>= 3.0.0)
RoxygenNote: 7.3.1
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-15 14:32:26 UTC; tianxili
Author: Mingyu Qi [aut],
  Chen-Wei Hua [aut],
  Tianxi Li [aut, cre],
  Wen Zhou [aut]
Maintainer: Tianxi Li <tianxili@umn.edu>
Repository: CRAN
Date/Publication: 2026-04-21 18:22:07 UTC
Built: R 4.5.2; ; 2026-04-21 18:41:49 UTC; unix

# Generated by roxygen2: do not edit by hand

export(sparsify_moments)
export(subsample_moments)
export(two_sample_test)
importFrom(doParallel,registerDoParallel)
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(foreach,registerDoSEQ)
importFrom(parallel,detectCores)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(stats,complete.cases)
importFrom(stats,cov)
importFrom(stats,rbinom)
importFrom(stats,setNames)

# ============================================================
# Algorithm 2 -- Externally sparsified network moments.
# ============================================================


# ------------------------------------------------------------
# Internal workhorse (called from both the public wrapper and
# the Algorithm 3 subsampling loop).
# ------------------------------------------------------------

#' Compute externally sparsified network moments (internal)
#'
#' Uses \code{G1} to estimate the sparsification probability and
#' independently applies random edge removal to \code{G2} to reach the
#' target density \code{rho_target}.  Returns the density-normalised moment
#' vector \eqn{\bar\Psi_{\rho^\dagger}(G_1, G_2)}.
#'
#' @param G1 Dense binary adjacency matrix used as the density estimator.
#' @param G2 Dense binary adjacency matrix used as the moment source.
#' @param rho_target Numeric; target edge density \eqn{\rho^\dagger \in (0,1)}.
#' @param motif_specs Compiled motif specification from
#'   \code{\link{init_motif_specs}}.
#'
#' @return Named numeric vector of density-normalised moments
#'   \eqn{\hat\rho_{\tilde{G}_2}^{-|E(R)|} \cdot U_R(\tilde{G}_2)}, or a
#'   vector of \code{NA} values if the result is degenerate (zero density
#'   after edge removal).
#'
#' @importFrom stats setNames
#' @keywords internal
.sparsify_moments_internal <- function(G1, G2, rho_target, motif_specs) {

  m      <- length(motif_specs$names)
  na_vec <- stats::setNames(rep(NA_real_, m), motif_specs$names)

  k1      <- nrow(G1)
  rho_G1  <- sum(G1) / (k1 * (k1 - 1L))

  if (rho_G1 < 1e-9) return(na_vec)

  # Sparsification probability: retain each edge in G2 with this probability
  # so that G2's density is driven towards rho_target.
  p_hat <- min(1, rho_target / rho_G1)

  G2_sparsified <- sparsify_graph(G2, p_hat)
  stats         <- compute_network_moments(G2_sparsified, motif_specs)

  if (stats$rho < 1e-9) return(na_vec)

  # Density normalisation: Psi-bar = rho^{-|E(R)|} * U_R
  stats$rho^(-motif_specs$edge_counts) * stats$UR
}


# ------------------------------------------------------------
# Public wrapper
# ------------------------------------------------------------

#' Externally sparsified network moments (Algorithm 2)
#'
#' Uses one adjacency matrix (\code{G1}) to estimate the sparsification
#' probability and a second independent adjacency matrix (\code{G2}) to
#' compute density-normalised network moments after random edge removal to a
#' common target density.  The resulting statistic
#' \eqn{\bar\Psi_{\rho^\dagger}(G_1, G_2)} is the building block for the
#' two-sample test in \code{\link{two_sample_test}}.
#'
#' @section Algorithm:
#' \enumerate{
#'   \item Estimate \eqn{\hat{p} = \min\!\left(1,\,\rho^\dagger /
#'     \hat\rho_{G_1}\right)}.
#'   \item Randomly remove edges from \eqn{G_2}: each edge is retained
#'     independently with probability \eqn{\hat{p}}, yielding
#'     \eqn{\tilde{G}_2}.
#'   \item Return \eqn{\hat\rho_{\tilde{G}_2}^{-|E(R)|} \cdot
#'     U_R(\tilde{G}_2)} for each motif \eqn{R}.
#' }
#'
#' @param G1 A square symmetric binary adjacency matrix (dense or sparse
#'   \code{Matrix}) used to estimate the sparsification probability.
#' @param G2 A square symmetric binary adjacency matrix from which moments
#'   are computed after edge removal.
#' @param rho_target Numeric in \eqn{(0, 1)}; the target edge density
#'   \eqn{\rho^\dagger}.
#' @param motifs A list of motif specifications; see
#'   \code{\link{subsample_moments}} for the format.  Defaults to
#'   \code{list("v_shape", "triangle", "star3")}.
#'
#' @return A named numeric vector of density-normalised network moments, or
#'   \code{NA} values if the graph after edge removal has zero density.
#'
#' @references
#' Qi, Hua, Li and Zhou (2024). Multivariate Inference of Network Moments by
#' Subsampling. \emph{Biometrika}, \strong{111}(1), 1--18.
#' \doi{10.1093/biomet/asad056}.
#'
#' @examples
#' \donttest{
#' library(randnet)
#' set.seed(2)
#' G1 <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A
#' G2 <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A
#' sparsify_moments(G1, G2, rho_target = 0.08)
#' }
#'
#' @export
sparsify_moments <- function(
    G1,
    G2,
    rho_target,
    motifs = list("v_shape", "triangle", "star3")
) {
  validate_adjacency(G1, "G1")
  validate_adjacency(G2, "G2")

  if (!is.numeric(rho_target) || length(rho_target) != 1L ||
      rho_target <= 0 || rho_target >= 1) {
    stop("'rho_target' must be a single numeric value strictly in (0, 1).")
  }

  motif_specs <- init_motif_specs(motifs)

  .sparsify_moments_internal(
    as.matrix(G1), as.matrix(G2), rho_target, motif_specs
  )
}

test_that("init_motif_specs accepts built-in names and produces correct structure", {
  specs <- init_motif_specs(list("v_shape", "triangle", "star3"))

  expect_equal(specs$names, c("v_shape", "triangle", "star3"))
  expect_equal(specs$edge_counts, c(2L, 3L, 3L))
  expect_true(is.function(specs$compute))
})

test_that("init_motif_specs rejects unknown built-in names", {
  expect_error(
    init_motif_specs(list("pentagon")),
    regexp = "Unknown built-in motif"
  )
})

test_that("init_motif_specs accepts a user-supplied spec returning one value", {
  my_spec <- list(
    fn      = function(A) c(my_edge = sum(A) / 2),
    n_edges = 1L
  )
  specs <- init_motif_specs(list(my_spec))

  expect_equal(specs$names, "my_edge")
  expect_equal(specs$edge_counts, 1L)
  A <- make_adj(10, 0.3, seed = 1)
  out <- specs$compute(A)
  expect_equal(out[["my_edge"]], sum(A) / 2)
})

test_that("init_motif_specs accepts a user-supplied spec returning two values", {
  my_spec <- list(
    fn      = function(A) c(motif_a = sum(A) / 2, motif_b = nrow(A)),
    n_edges = c(1L, 0L)
  )
  specs <- init_motif_specs(list("v_shape", my_spec))

  expect_equal(specs$names, c("v_shape", "motif_a", "motif_b"))
  expect_equal(length(specs$edge_counts), 3L)
})

test_that("init_motif_specs errors when fn output length mismatches n_edges", {
  bad_spec <- list(
    fn      = function(A) c(a = 1, b = 2),
    n_edges = 3L  # wrong length
  )
  expect_error(
    init_motif_specs(list(bad_spec)),
    regexp = "Length of 'n_edges'"
  )
})

test_that("init_motif_specs errors when fn is not a function", {
  bad_spec <- list(fn = 42, n_edges = 1L)
  expect_error(init_motif_specs(list(bad_spec)), regexp = "must be a function")
})

# ------------------------------------------------------------------

test_that("compute_network_moments returns correct density", {
  A     <- make_adj(10, 0.4, seed = 5)
  specs <- init_motif_specs(list("v_shape"))
  out   <- compute_network_moments(A, specs)

  expected_rho <- sum(A) / (10 * 9)
  expect_equal(out$rho, expected_rho)
})

test_that("compute_network_moments returns zeros for empty graph", {
  A     <- matrix(0L, 8, 8)
  specs <- init_motif_specs(list("v_shape", "triangle", "star3"))
  out   <- compute_network_moments(A, specs)

  expect_equal(out$rho, 0)
  expect_true(all(out$UR == 0))
})

test_that("compute_network_moments returns zeros when k < 3", {
  A     <- make_adj(2, 0.5, seed = 1)
  specs <- init_motif_specs(list("v_shape", "triangle"))
  out   <- compute_network_moments(A, specs)

  expect_equal(out$rho, 0)
  expect_true(all(out$UR == 0))
})

test_that("compute_network_moments v_shape moment is non-negative", {
  A     <- make_adj(20, 0.3, seed = 7)
  specs <- init_motif_specs(list("v_shape"))
  out   <- compute_network_moments(A, specs)
  expect_true(out$UR[["v_shape"]] >= 0)
})

# ------------------------------------------------------------------

test_that("sparsify_graph returns matrix of same dimensions", {
  A    <- make_adj(15, 0.5, seed = 2)
  out  <- sparsify_graph(A, 0.5)
  expect_equal(dim(out), dim(A))
})

test_that("sparsify_graph with p=1 is identity", {
  A   <- make_adj(10, 0.4, seed = 3)
  out <- sparsify_graph(A, 1)
  expect_identical(out, A)
})

test_that("sparsify_graph with p=0 gives zero matrix", {
  A   <- make_adj(10, 0.4, seed = 4)
  out <- sparsify_graph(A, 0)
  expect_true(all(out == 0L))
})

test_that("sparsify_graph result is symmetric", {
  set.seed(10)
  A   <- make_adj(12, 0.5, seed = 10)
  out <- sparsify_graph(A, 0.6)
  expect_equal(out, t(out))
})

test_that("sparsify_graph result is a subset of input edges", {
  A   <- make_adj(20, 0.6, seed = 11)
  out <- sparsify_graph(A, 0.4)
  # Every edge in out must also be in A.
  expect_true(all(out[A == 0L] == 0L))
})

# ------------------------------------------------------------------

test_that("subsample_moments returns correct structure", {
  A   <- make_adj(80, 0.15, seed = 20)
  res <- subsample_moments(A, b = 20, N_sub = 30)

  expect_named(res, c("rho", "moments", "motif_names", "b"))
  expect_equal(nrow(res$moments), 30L)
  expect_equal(ncol(res$moments), 3L)
  expect_equal(res$motif_names, c("v_shape", "triangle", "star3"))
  expect_equal(res$b, 20L)
})

test_that("subsample_moments density estimate is in (0,1)", {
  A   <- make_adj(60, 0.2, seed = 21)
  res <- subsample_moments(A, b = 15, N_sub = 20)
  expect_true(res$rho > 0 && res$rho < 1)
})

test_that("subsample_moments errors when b >= n", {
  A <- make_adj(20, 0.2, seed = 22)
  expect_error(subsample_moments(A, b = 20, N_sub = 10), regexp = "strictly less")
})

test_that("subsample_moments works with a single motif", {
  A   <- make_adj(50, 0.2, seed = 23)
  res <- subsample_moments(A, b = 15, N_sub = 20, motifs = list("triangle"))

  expect_equal(ncol(res$moments), 1L)
  expect_equal(colnames(res$moments), "triangle")
})

test_that("subsample_moments works with a user-supplied motif", {
  edge_count_fn <- list(
    fn      = function(A) c(n_edges = sum(A) / 2),
    n_edges = 1L
  )
  A   <- make_adj(50, 0.2, seed = 24)
  res <- subsample_moments(A, b = 15, N_sub = 20,
                           motifs = list("v_shape", edge_count_fn))

  expect_equal(ncol(res$moments), 2L)
  expect_equal(colnames(res$moments), c("v_shape", "n_edges"))
})

# ------------------------------------------------------------------

test_that("sparsify_moments returns named vector of correct length", {
  G1  <- make_adj(40, 0.20, seed = 30)
  G2  <- make_adj(40, 0.20, seed = 31)
  out <- sparsify_moments(G1, G2, rho_target = 0.08)

  expect_true(is.numeric(out))
  expect_equal(length(out), 3L)
  expect_equal(names(out), c("v_shape", "triangle", "star3"))
})

test_that("sparsify_moments returns NA for empty G1", {
  G1  <- matrix(0L, 20, 20)
  G2  <- make_adj(20, 0.3, seed = 32)
  out <- sparsify_moments(G1, G2, rho_target = 0.05)
  expect_true(all(is.na(out)))
})

test_that("sparsify_moments errors on invalid rho_target", {
  G1 <- make_adj(20, 0.2, seed = 33)
  G2 <- make_adj(20, 0.2, seed = 34)
  expect_error(sparsify_moments(G1, G2, rho_target = 1.5))
  expect_error(sparsify_moments(G1, G2, rho_target = 0))
})

# ------------------------------------------------------------------

test_that("two_sample_test returns correct structure (mahalanobis)", {
  set.seed(40)
  G       <- make_adj(200, 0.10)
  G_prime <- make_adj(60,  0.10)

  res <- two_sample_test(G, G_prime, rho_target = 0.06, N_sub = 100)

  expect_named(
    res,
    c("p_value", "test_statistic", "observed_moments", "null_moments",
      "null_statistic", "method", "motif_names")
  )
  expect_true(res$p_value >= 0 && res$p_value <= 1)
  expect_equal(res$method, "mahalanobis")
  expect_equal(res$motif_names, c("v_shape", "triangle", "star3"))
  expect_equal(nrow(res$null_moments), 100L)
})

test_that("two_sample_test returns correct structure (cauchy)", {
  set.seed(41)
  G       <- make_adj(200, 0.10)
  G_prime <- make_adj(60,  0.10)

  res <- two_sample_test(G, G_prime, rho_target = 0.06,
                         N_sub = 100, test = "cauchy")

  expect_equal(res$method, "cauchy")
  expect_named(res$marginal_p_values, c("v_shape", "triangle", "star3"))
  expect_true(res$p_value >= 0 && res$p_value <= 1)
})

test_that("two_sample_test p-value is near nominal level under null (rough check)", {
  # Generate 20 repetitions under the null; expect rejection rate near 0.05
  # for alpha = 0.05.  With 20 reps this is a very loose sanity check.
  set.seed(42)
  pvals <- replicate(20, {
    G       <- make_adj(150, 0.12)
    G_prime <- make_adj(50,  0.12)
    res     <- two_sample_test(G, G_prime, rho_target = 0.07,
                               N_sub = 200, test = "mahalanobis")
    res$p_value
  })
  # Under the null, p-values should be roughly uniform, so most > 0.05.
  expect_true(mean(pvals > 0.05) > 0.5)
})

test_that("two_sample_test errors when G_prime too small to split", {
  G       <- make_adj(50, 0.2, seed = 50)
  G_prime <- make_adj(4,  0.2, seed = 51)   # b_half = 2 < 3
  expect_error(
    two_sample_test(G, G_prime, rho_target = 0.1, N_sub = 50),
    regexp = "at least 6"
  )
})

test_that("two_sample_test errors on invalid rho_target", {
  G       <- make_adj(100, 0.1, seed = 52)
  G_prime <- make_adj(30,  0.1, seed = 53)
  expect_error(two_sample_test(G, G_prime, rho_target = 2,  N_sub = 50))
  expect_error(two_sample_test(G, G_prime, rho_target = -1, N_sub = 50))
})

test_that("two_sample_test validate_adjacency catches non-square input", {
  bad <- matrix(0L, 3, 4)
  G   <- make_adj(50, 0.1, seed = 54)
  expect_error(two_sample_test(bad, G, rho_target = 0.05, N_sub = 50))
})

# ============================================================
# Algorithm 3 -- Subsampling comparison between unmatchable
#                networks with unequal edge densities.
# ============================================================


#' Two-sample test for unmatchable networks with unequal densities (Algorithm 3)
#'
#' Tests whether two unmatchable networks --- differing in node set, size, and
#' possibly edge density --- arise from the same underlying graphon model.
#' The procedure avoids the density-mismatch problem via a
#' \emph{split-and-sparsify} strategy: each network is randomly split into two
#' halves, one used to estimate the sparsification probability and the other to
#' compute density-normalised moments at a shared target density.  Node
#' subsampling from the larger network then generates a reference distribution
#' against which the observed statistic is compared.
#'
#' @section Algorithm:
#' \enumerate{
#'   \item Randomly partition the nodes of \code{G_prime} into two equal
#'     halves \eqn{G'_1} and \eqn{G'_2}; compute the observed statistic
#'     \eqn{\bar\Psi_{\rho^\dagger}(G'_1, G'_2)} via Algorithm 2.
#'   \item For \eqn{i = 1, \ldots, N_{\mathrm{sub}}}: independently draw
#'     two sets of \eqn{\lfloor b/2 \rfloor} nodes from \eqn{G} and compute
#'     \eqn{\bar\Psi_{\rho^\dagger}(G^*_{i1}, G^*_{i2})} via Algorithm 2.
#'   \item Compare the observed statistic against the empirical null
#'     distribution using the selected \code{test}.
#' }
#'
#' @section Choice of \code{rho_target}:
#' A recommended default is
#' \eqn{\rho^\dagger = \kappa \cdot \min(\hat\rho_G, \hat\rho_{G'})},
#' with \eqn{\kappa \in (0.5,\, 0.9)}.  Values close to 0 discard too many
#' edges and inflate variance; values too close to
#' \eqn{\min(\hat\rho_G, \hat\rho_{G'})} leave little room for edge removal
#' and can break the density-matching property.  A value of \eqn{\kappa = 0.7}
#' (corresponding to
#' \code{rho_target = 0.7 * min(rho_hat_G, rho_hat_G_prime)}) works well
#' across a range of settings.
#'
#' @section Tests:
#' \describe{
#'   \item{Mahalanobis (default)}{Computes a multivariate distance that fully
#'     exploits the joint distribution of all moments.  The p-value is the
#'     proportion of null distances exceeding the observed distance.  This test
#'     is more powerful when the signal lies in the \emph{joint} rather than
#'     the \emph{marginal} distribution of the moments.}
#'   \item{Cauchy combination}{Combines marginal p-values via the Cauchy
#'     combination test of Liu & Xie (2020).  Controls type I error under
#'     arbitrary dependence but ignores cross-moment correlation.}
#' }
#'
#' @param G A square symmetric binary adjacency matrix (dense or sparse
#'   \code{Matrix}) of size \eqn{n \times n}.  Typically the larger or
#'   reference network.
#' @param G_prime A square symmetric binary adjacency matrix of size
#'   \eqn{b \times b}.  The network to compare against \code{G}.
#' @param rho_target Numeric in \eqn{(0, 1)}; the common target edge density
#'   \eqn{\rho^\dagger}.  See the \emph{Choice of \code{rho_target}} section.
#' @param N_sub Integer; number of subsampling replicates.  Default 2000.
#' @param motifs A list of motif specifications; see
#'   \code{\link{subsample_moments}} for the format.  Defaults to
#'   \code{list("v_shape", "triangle", "star3")}.
#' @param test Character; the multivariate test to apply.  Either
#'   \code{"mahalanobis"} (default) or \code{"cauchy"}.
#' @param parallel Logical; whether to parallelise the subsampling loop.
#'   Default \code{FALSE}.  Requires the \pkg{netsubsamp} package to be
#'   \emph{installed} on all worker nodes when \code{TRUE}.
#' @param n_cores Integer or \code{NULL}; number of cores when
#'   \code{parallel = TRUE}.
#'
#' @return A list with:
#' \describe{
#'   \item{\code{p_value}}{The p-value of the test.}
#'   \item{\code{test_statistic}}{Observed test statistic: Mahalanobis
#'     distance \eqn{D_0} when \code{test = "mahalanobis"}, or Cauchy
#'     combination statistic \eqn{T} when \code{test = "cauchy"}.}
#'   \item{\code{observed_moments}}{Named numeric vector; the observed
#'     \eqn{\bar\Psi} statistic.}
#'   \item{\code{null_moments}}{\eqn{N_{\mathrm{sub}} \times m} matrix of
#'     null \eqn{\bar\Psi} statistics (including any \code{NA} rows from
#'     degenerate subsamples).}
#'   \item{\code{null_statistic}}{Numeric vector of per-replicate null
#'     Mahalanobis distances (or marginal p-values for \code{"cauchy"}).}
#'   \item{\code{marginal_p_values}}{(Only for \code{test = "cauchy"}.) Named
#'     numeric vector of per-motif marginal p-values.}
#'   \item{\code{method}}{The test method used.}
#'   \item{\code{motif_names}}{Character vector of moment names.}
#' }
#'
#' @references
#' Qi, Hua, Li and Zhou (2024). Multivariate Inference of Network Moments by
#' Subsampling. \emph{Biometrika}, \strong{111}(1), 1--18.
#' \doi{10.1093/biomet/asad056}.
#'
#' Liu, Y. and Xie, J. (2020). Cauchy combination test: a powerful test with
#' analytic p-value calculation under arbitrary dependency structures.
#' \emph{Journal of the American Statistical Association}, \strong{115},
#' 393--402. \doi{10.1080/01621459.2018.1554485}.
#'
#' @examples
#' \donttest{
#' library(randnet)
#' set.seed(3)
#' G       <- BlockModel.Gen(lambda = 20, n = 400, K = 3)$A
#' G_prime <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A
#'
#' # Choose rho_target as 0.7 * min(density(G), density(G_prime))
#' rho_G       <- sum(G)       / (400 * 399)
#' rho_G_prime <- sum(G_prime) / (100 * 99)
#' rho_target  <- 0.7 * min(rho_G, rho_G_prime)
#'
#' res <- two_sample_test(G, G_prime,
#'                        rho_target = rho_target,
#'                        N_sub      = 500,
#'                        test       = "mahalanobis")
#' res$p_value
#' }
#'
#' @importFrom foreach foreach "%dopar%"
#' @importFrom stats complete.cases cov setNames
#' @export
two_sample_test <- function(
    G,
    G_prime,
    rho_target,
    N_sub    = 2000L,
    motifs   = list("v_shape", "triangle", "star3"),
    test     = c("mahalanobis", "cauchy"),
    parallel = FALSE,
    n_cores  = NULL
) {
  test  <- match.arg(test)
  validate_adjacency(G,       "G")
  validate_adjacency(G_prime, "G_prime")

  n     <- nrow(G)
  b     <- nrow(G_prime)
  N_sub <- as.integer(N_sub)

  if (b >= n) {
    warning(
      "'G_prime' is not smaller than 'G'; results may be unreliable. ",
      "Algorithm 3 is designed for the case nrow(G_prime) < nrow(G)."
    )
  }

  b_half <- floor(b / 2L)
  if (b_half < 3L) {
    stop("nrow(G_prime) must be at least 6 to allow an equal split.")
  }
  if (N_sub < 10L) {
    stop("'N_sub' must be at least 10.")
  }
  if (!is.numeric(rho_target) || length(rho_target) != 1L ||
      rho_target <= 0 || rho_target >= 1) {
    stop("'rho_target' must be a single numeric value strictly in (0, 1).")
  }

  motif_specs <- init_motif_specs(motifs)
  m           <- length(motif_specs$names)

  # ------------------------------------------------------------------
  # Step 1: Split G_prime and compute the observed statistic.
  # ------------------------------------------------------------------
  G_prime_mat <- as.matrix(G_prime)
  all_nodes   <- seq_len(b)
  half1_idx   <- sample(all_nodes, b_half, replace = FALSE)
  half2_idx   <- setdiff(all_nodes, half1_idx)

  G_prime_1 <- G_prime_mat[half1_idx, half1_idx]
  G_prime_2 <- G_prime_mat[half2_idx, half2_idx]

  psi_obs <- .sparsify_moments_internal(
    G_prime_1, G_prime_2, rho_target, motif_specs
  )

  if (anyNA(psi_obs)) {
    warning(
      "The observed statistic contains NA (zero density after edge removal). ",
      "Consider using a smaller 'rho_target' or a larger 'G_prime'."
    )
  }

  # ------------------------------------------------------------------
  # Step 2: Subsampling loop -- generate null distribution.
  # ------------------------------------------------------------------
  par_setup <- setup_foreach(parallel, n_cores)
  on.exit(par_setup$cleanup(), add = TRUE)
  `%op%` <- par_setup$op

  G_local          <- G
  b_half_local     <- b_half
  rho_target_local <- rho_target
  specs_local      <- motif_specs

  pkg_arg <- if (parallel) "netsubsamp" else character(0L)

  null_list <- foreach::foreach(
    i         = seq_len(N_sub),
    .packages = pkg_arg,
    .export   = c(".sparsify_moments_internal", "sparsify_graph",
                  "compute_network_moments")
  ) %op% {
    n_w    <- nrow(G_local)
    idx1   <- sample(n_w, b_half_local, replace = FALSE)
    idx2   <- sample(n_w, b_half_local, replace = FALSE)
    G_sub1 <- as.matrix(G_local[idx1, idx1])
    G_sub2 <- as.matrix(G_local[idx2, idx2])
    .sparsify_moments_internal(G_sub1, G_sub2, rho_target_local, specs_local)
  }

  null_mat           <- do.call(rbind, null_list)
  colnames(null_mat) <- motif_specs$names

  # Rows with NA arise from degenerate (zero-density) subsamples.
  valid       <- stats::complete.cases(null_mat)
  null_valid  <- null_mat[valid, , drop = FALSE]
  n_valid     <- nrow(null_valid)

  if (n_valid < 10L) {
    stop(
      "Fewer than 10 valid null replicates obtained (", n_valid, " of ",
      N_sub, "). Consider a larger 'N_sub', a larger 'G', or a smaller ",
      "'rho_target'."
    )
  }

  # ------------------------------------------------------------------
  # Step 3: Compute test statistic and p-value.
  # ------------------------------------------------------------------
  eta_hat <- colMeans(null_valid)

  if (test == "mahalanobis") {

    Sigma_hat <- stats::cov(null_valid)

    Sigma_inv <- tryCatch(
      solve(Sigma_hat),
      error = function(e) {
        warning(
          "Covariance matrix is singular or near-singular; ",
          "applying ridge regularisation (lambda = 1e-6 * trace / m)."
        )
        lambda <- 1e-6 * sum(diag(Sigma_hat)) / m
        solve(Sigma_hat + diag(lambda, m))
      }
    )

    .mahal_dist <- function(x) {
      d <- x - eta_hat
      as.numeric(crossprod(d, Sigma_inv %*% d))
    }

    D_null <- apply(null_valid, 1, .mahal_dist)
    D_obs  <- .mahal_dist(psi_obs)

    p_value <- mean(D_null > D_obs, na.rm = TRUE)

    list(
      p_value          = p_value,
      test_statistic   = D_obs,
      observed_moments = psi_obs,
      null_moments     = null_mat,
      null_statistic   = D_null,
      method           = "mahalanobis",
      motif_names      = motif_specs$names
    )

  } else {
    # Cauchy combination test (Liu & Xie, 2020).
    obs_dev  <- abs(psi_obs  - eta_hat)

    null_dev <- abs(sweep(null_valid, 2L, eta_hat, FUN = "-"))

    # Per-motif marginal p-values: proportion of null |deviations| that
    # exceed the observed |deviation|.
    p_marginal <- vapply(seq_len(m), function(j) {
      mean(null_dev[, j] > obs_dev[j], na.rm = TRUE)
    }, numeric(1L))
    names(p_marginal) <- motif_specs$names

    # Cauchy statistic and combined p-value.
    T_cauchy <- mean(tan((0.5 - p_marginal) * pi))
    p_cauchy <- 0.5 - atan(T_cauchy) / pi

    list(
      p_value           = p_cauchy,
      test_statistic    = T_cauchy,
      marginal_p_values = p_marginal,
      observed_moments  = psi_obs,
      null_moments      = null_mat,
      null_statistic    = p_marginal,
      method            = "cauchy",
      motif_names       = motif_specs$names
    )
  }
}

# ============================================================
# Internal helper: compute edge density and network moments
# for a (small, dense) adjacency matrix.
# ============================================================


#' Compute edge density and network moments
#'
#' Given a square binary adjacency matrix and a compiled motif specification,
#' returns the edge density and the vector of network moments (rescaled motif
#' counts).
#'
#' For a network with \eqn{k} nodes the edge density is
#' \eqn{\hat\rho = |E| / [k(k-1)]}.  The network moment for motif \eqn{R}
#' with \eqn{r} nodes is \eqn{U_R = \binom{k}{r}^{-1} X_R}, where \eqn{X_R}
#' is the (non-induced) motif count.  Built-in moments use matrix-algebraic
#' formulas that avoid explicit enumeration.
#'
#' @param A A square binary adjacency matrix (\strong{dense}).  Must have at
#'   least 3 rows/columns for V-shape and triangle moments, and at least 4 for
#'   the 3-star moment.
#' @param motif_specs Compiled motif specification, as returned by
#'   \code{\link{init_motif_specs}}.
#'
#' @return A list with:
#' \describe{
#'   \item{\code{rho}}{Scalar edge density.}
#'   \item{\code{UR}}{Named numeric vector of network moments (one per motif
#'     in \code{motif_specs}).}
#' }
#' Rows corresponding to graphs with zero density return \code{rho = 0} and
#' all moments equal to zero.
#'
#' @importFrom stats setNames
#' @keywords internal
compute_network_moments <- function(A, motif_specs) {

  k   <- nrow(A)
  rho <- sum(A) / (k * (k - 1L))

  m        <- length(motif_specs$names)
  zero_vec <- stats::setNames(rep(0, m), motif_specs$names)

  if (rho < 1e-9 || k < 3L) {
    return(list(rho = 0, UR = zero_vec))
  }

  list(rho = rho, UR = motif_specs$compute(A))
}

# ============================================================
# Algorithm 1 -- Uniform node subsampling for multivariate
#                network moments.
# ============================================================


#' Uniform node subsampling for multivariate network moments (Algorithm 1)
#'
#' Repeatedly draws random node-induced subgraphs of size \code{b} from
#' network \code{G} and computes network moments for each subsample.  The
#' resulting empirical distribution approximates the joint distribution of
#' network moments for networks of size \code{b} drawn from the same
#' underlying graphon model, providing the foundation for downstream inference.
#'
#' @param G A square symmetric binary adjacency matrix (dense \code{matrix}
#'   or sparse \code{Matrix}) of size \eqn{n \times n}.
#' @param b Integer; the subsampling size.  Must satisfy \eqn{3 \le b < n}.
#' @param N_sub Integer; number of subsamples to draw.
#' @param motifs A list of motif specifications.  Each element is either:
#'   \itemize{
#'     \item a character string: \code{"v_shape"}, \code{"triangle"}, or
#'       \code{"star3"} (built-in defaults); or
#'     \item a named list with elements \code{fn} -- a
#'       \code{function(A)} returning a \emph{named} numeric vector -- and
#'       \code{n_edges} -- an integer vector of edge counts matching the
#'       output length of \code{fn}.
#'   }
#'   Defaults to \code{list("v_shape", "triangle", "star3")}.
#' @param parallel Logical; if \code{TRUE} the subsampling loop is run in
#'   parallel via \pkg{doParallel} and \pkg{foreach}.  Default \code{FALSE}.
#'   Requires the \pkg{netsubsamp} package to be \emph{installed} on all
#'   worker nodes when \code{TRUE}.
#' @param n_cores Integer or \code{NULL}; number of cores to use when
#'   \code{parallel = TRUE}.  If \code{NULL}, uses
#'   \code{parallel::detectCores() - 1} (minimum 1).
#'
#' @return A list with:
#' \describe{
#'   \item{\code{rho}}{Estimated edge density of \code{G}.}
#'   \item{\code{moments}}{An \eqn{N_{\mathrm{sub}} \times m} numeric matrix
#'     of subsampled network moments.  Rows for which the subsampled graph has
#'     zero density contain \code{NA}.}
#'   \item{\code{motif_names}}{Character vector of moment names, in column
#'     order of \code{moments}.}
#'   \item{\code{b}}{The subsampling size used.}
#' }
#'
#' @references
#' Qi, Hua, Li and Zhou (2024). Multivariate Inference of Network Moments by
#' Subsampling. \emph{Biometrika}, \strong{111}(1), 1--18.
#' \doi{10.1093/biomet/asad056}.
#'
#' @examples
#' \donttest{
#' library(randnet)
#' set.seed(1)
#' G   <- BlockModel.Gen(lambda = 15, n = 200, K = 3)$A
#' res <- subsample_moments(G, b = 60, N_sub = 500)
#' head(res$moments)
#' res$rho
#' }
#'
#' @importFrom foreach foreach "%dopar%"
#' @export
subsample_moments <- function(
    G,
    b,
    N_sub,
    motifs   = list("v_shape", "triangle", "star3"),
    parallel = FALSE,
    n_cores  = NULL
) {
  validate_adjacency(G, "G")
  n     <- nrow(G)
  b     <- as.integer(b)
  N_sub <- as.integer(N_sub)

  if (b >= n)   stop("'b' must be strictly less than nrow(G).")
  if (b < 3L)   stop("'b' must be at least 3.")
  if (N_sub < 1L) stop("'N_sub' must be a positive integer.")

  motif_specs <- init_motif_specs(motifs)
  m           <- length(motif_specs$names)

  par_setup <- setup_foreach(parallel, n_cores)
  on.exit(par_setup$cleanup(), add = TRUE)
  `%op%` <- par_setup$op

  # Freeze all variables needed inside the loop to avoid accidental
  # re-evaluation or masking in the parallel environment.
  G_local      <- G
  b_local      <- b
  specs_local  <- motif_specs

  pkg_arg <- if (parallel) "netsubsamp" else character(0L)

  result_list <- foreach::foreach(
    i         = seq_len(N_sub),
    .packages = pkg_arg,
    .export   = c("compute_network_moments", "init_motif_specs")
  ) %op% {
    n_w    <- nrow(G_local)
    nodes  <- sample(n_w, b_local, replace = FALSE)
    G_sub  <- as.matrix(G_local[nodes, nodes])

    stats  <- compute_network_moments(G_sub, specs_local)

    if (stats$rho < 1e-9) {
      rep(NA_real_, length(specs_local$names))
    } else {
      stats$UR
    }
  }

  moments_mat <- do.call(rbind, result_list)
  rownames(moments_mat) <- NULL
  colnames(moments_mat) <- motif_specs$names

  rho_G <- sum(G) / (n * (n - 1L))

  list(
    rho         = rho_G,
    moments     = moments_mat,
    motif_names = motif_specs$names,
    b           = b
  )
}

# ============================================================
# Internal utilities: motif spec initialisation, input
# validation, and parallel backend management.
# ============================================================


# ------------------------------------------------------------
# Motif specification
# ------------------------------------------------------------

#' Initialise and validate a list of motif specifications
#'
#' Parses a mixed list of built-in motif names (character strings) and
#' user-supplied specifications (lists with \code{fn} and \code{n_edges})
#' into an internal representation used throughout the package.
#'
#' Built-in motifs and their edge counts:
#' \itemize{
#'   \item \code{"v_shape"} -- 2-star (V-shape), 2 edges
#'   \item \code{"triangle"} -- triangle, 3 edges
#'   \item \code{"star3"} -- 3-star, 3 edges
#' }
#'
#' A user-supplied specification must be a named list with:
#' \itemize{
#'   \item \code{fn}: a \code{function(A)} that accepts a square binary
#'     adjacency matrix and returns a \emph{named} numeric vector of one or
#'     more moment values (one element per motif computed by that function).
#'   \item \code{n_edges}: integer vector whose length equals the length of
#'     \code{fn}'s return value, giving the number of edges in each
#'     corresponding motif.  Used for density normalisation.
#' }
#'
#' @param motifs A list of motif specifications (see above).
#'
#' @return A list with three elements:
#' \describe{
#'   \item{\code{compute}}{A function \code{function(A)} that evaluates all
#'     motifs on an adjacency matrix and returns a named numeric vector.}
#'   \item{\code{edge_counts}}{Integer vector of edge counts, one per moment.}
#'   \item{\code{names}}{Character vector of moment names, in the same order
#'     as \code{edge_counts}.}
#' }
#'
#' @importFrom stats setNames
#' @keywords internal
init_motif_specs <- function(motifs) {

  builtin_edge_counts <- c(v_shape = 2L, triangle = 3L, star3 = 3L)

  all_names       <- character(0)
  all_edge_counts <- integer(0)
  fn_list         <- list()

  # Small connected graph used to probe user-supplied functions.
  dummy_A <- matrix(
    c(0L, 1L, 1L, 0L, 0L,
      1L, 0L, 1L, 1L, 0L,
      1L, 1L, 0L, 0L, 1L,
      0L, 1L, 0L, 0L, 1L,
      0L, 0L, 1L, 1L, 0L),
    nrow = 5L, ncol = 5L
  )

  for (spec in motifs) {

    if (is.character(spec)) {
      # --- built-in motif ---
      if (!spec %in% names(builtin_edge_counts)) {
        stop(sprintf(
          "Unknown built-in motif '%s'. Valid names: %s",
          spec, paste(names(builtin_edge_counts), collapse = ", ")
        ))
      }
      all_names       <- c(all_names,       spec)
      all_edge_counts <- c(all_edge_counts, builtin_edge_counts[[spec]])
      fn_list         <- c(fn_list, list(list(type = "builtin", name = spec)))

    } else if (is.list(spec) && !is.null(spec$fn) && !is.null(spec$n_edges)) {
      # --- user-supplied spec ---
      if (!is.function(spec$fn)) {
        stop("'fn' in a user-supplied motif specification must be a function.")
      }

      # Call fn on dummy matrix to discover output names and validate length.
      test_out <- tryCatch(
        spec$fn(dummy_A),
        error = function(e) stop(
          "User-supplied motif function failed on a 5x5 test matrix: ",
          conditionMessage(e)
        )
      )
      if (!is.numeric(test_out)) {
        stop("User-supplied motif function must return a numeric vector.")
      }
      n_out    <- length(test_out)
      n_edges  <- as.integer(spec$n_edges)
      if (length(n_edges) != n_out) {
        stop(sprintf(
          "Length of 'n_edges' (%d) must match the output length of 'fn' (%d).",
          length(n_edges), n_out
        ))
      }

      # Assign names: use names from fn output, or generate defaults.
      custom_names <- if (!is.null(names(test_out))) {
        names(test_out)
      } else {
        paste0("custom_", seq_len(n_out))
      }

      all_names       <- c(all_names,       custom_names)
      all_edge_counts <- c(all_edge_counts, n_edges)
      fn_list         <- c(fn_list, list(list(type = "custom", fn = spec$fn)))

    } else {
      stop(
        "Each motif specification must be a character string or a list with ",
        "elements 'fn' (a function) and 'n_edges' (an integer vector)."
      )
    }
  }

  if (length(all_names) == 0L) {
    stop("'motifs' must contain at least one motif specification.")
  }

  # Determine once whether A^2 is needed (shared by v_shape and triangle).
  builtin_names <- vapply(
    fn_list,
    function(x) if (x$type == "builtin") x$name else "",
    character(1L)
  )
  needs_A2 <- any(builtin_names %in% c("v_shape", "triangle"))

  # Build the combined compute function.  All captured objects are frozen in
  # a local environment so the closure remains self-contained.
  compute_fn <- local({
    fns     <- fn_list
    use_A2  <- needs_A2
    nms     <- all_names

    function(A) {
      k      <- nrow(A)
      result <- numeric(0L)
      A2     <- if (use_A2) A %*% A else NULL

      for (spec in fns) {
        if (spec$type == "builtin") {
          val <- switch(
            spec$name,
            v_shape = {
              if (k < 3L) {
                0
              } else {
                count_v <- sum(A2) - sum(diag(A2))
                count_v / (k * (k - 1L) * (k - 2L) / 3L)
              }
            },
            triangle = {
              if (k < 3L) {
                0
              } else {
                sum(A * A2) / (k * (k - 1L) * (k - 2L))
              }
            },
            star3 = {
              if (k < 4L) {
                0
              } else {
                vv <- colSums(A)
                sum(choose(vv, 3L)) /
                  (k * (k - 1L) * (k - 2L) * (k - 3L) / 24L)
              }
            }
          )
          result <- c(result, stats::setNames(val, spec$name))
        } else {
          result <- c(result, spec$fn(A))
        }
      }
      result
    }
  })

  list(
    compute     = compute_fn,
    edge_counts = all_edge_counts,
    names       = all_names
  )
}


# ------------------------------------------------------------
# Input validation
# ------------------------------------------------------------

#' Validate an adjacency matrix
#'
#' Checks that \code{A} is a square matrix (dense or sparse \code{Matrix})
#' and warns if it does not appear symmetric.
#'
#' @param A Object to validate.
#' @param name Name used in error messages.
#'
#' @keywords internal
validate_adjacency <- function(A, name = "A") {
  if (!is.matrix(A) && !inherits(A, "Matrix")) {
    stop(sprintf("'%s' must be a matrix or sparse Matrix object.", name))
  }
  if (nrow(A) != ncol(A)) {
    stop(sprintf(
      "'%s' must be square (got %d x %d).",
      name, nrow(A), ncol(A)
    ))
  }
  if (is.matrix(A) && max(abs(A - t(A))) > 1e-10) {
    warning(sprintf(
      "'%s' does not appear to be symmetric; results may be unreliable.", name
    ))
  }
}


# ------------------------------------------------------------
# Parallel backend management
# ------------------------------------------------------------

#' Set up a foreach parallel backend
#'
#' Registers a \pkg{doParallel} backend when requested and returns the
#' appropriate foreach binary operator together with a cleanup function.
#'
#' @param parallel Logical; whether parallelism is requested.
#' @param n_cores  Integer or \code{NULL}.  When \code{parallel = TRUE} and
#'   \code{n_cores} is \code{NULL}, one fewer than the number of detected
#'   cores is used (minimum 1).
#'
#' @return A list with:
#' \describe{
#'   \item{\code{op}}{The foreach binary operator (\code{\%do\%} or
#'     \code{\%dopar\%}).}
#'   \item{\code{cleanup}}{A zero-argument function that stops any cluster
#'     created here and re-registers the sequential backend.}
#' }
#'
#' @importFrom doParallel registerDoParallel
#' @importFrom foreach registerDoSEQ
#' @importFrom parallel detectCores makeCluster stopCluster
#' @keywords internal
setup_foreach <- function(parallel, n_cores) {

  if (!parallel) {
    return(list(
      op      = foreach::`%do%`,
      cleanup = function() invisible(NULL)
    ))
  }

  cores <- if (!is.null(n_cores)) {
    max(1L, as.integer(n_cores))
  } else {
    max(1L, parallel::detectCores() - 1L)
  }

  cl <- parallel::makeCluster(cores)
  doParallel::registerDoParallel(cl)

  list(
    op      = foreach::`%dopar%`,
    cleanup = function() {
      parallel::stopCluster(cl)
      foreach::registerDoSEQ()
    }
  )
}

# ============================================================
# Package-level documentation.
# ============================================================

#' netsubsamp: Multivariate Inference of Network Moments by Subsampling
#'
#' @description
#' \pkg{netsubsamp} implements the three subsampling algorithms from
#' Qi, Hua, Li and Zhou (2024) for multivariate inference on network moments
#' (rescaled counts of graph motifs such as V-shapes, triangles, and stars)
#' under a sparse graphon model.
#'
#' @section The three algorithms:
#'
#' \strong{Algorithm 1 -- \code{\link{subsample_moments}}: Uniform node
#' subsampling.}
#' Draws \eqn{N_{\mathrm{sub}}} random node-induced subgraphs of size
#' \eqn{b} from a single network \eqn{G} and computes the vector of network
#' moments for each subsample.  The resulting empirical distribution
#' consistently approximates the joint distribution of multiple network moments
#' for networks of size \eqn{b} drawn from the same graphon.  This provides
#' the foundation for any downstream inference task: confidence regions,
#' goodness-of-fit tests, or visualization of the joint moment distribution.
#'
#' \strong{Algorithm 2 -- \code{\link{sparsify_moments}}: Externally
#' sparsified moments.}
#' Given two independently sampled subgraphs \eqn{G_1} and \eqn{G_2} and a
#' target density \eqn{\rho^\dagger}, uses \eqn{G_1} to estimate the
#' required edge-removal probability and then randomly removes edges from
#' \eqn{G_2} to produce a graph at the target density.  Returns the
#' density-normalised moment vector
#' \eqn{\bar\Psi_{\rho^\dagger}(G_1, G_2)}.  This building block handles
#' the practical complication that the true graphon density is unknown.
#'
#' \strong{Algorithm 3 -- \code{\link{two_sample_test}}: Two-sample test for
#' unmatchable networks with unequal densities.}
#' Tests \eqn{H_0: w = w'} for two unmatchable networks \eqn{G} (size
#' \eqn{n}) and \eqn{G'} (size \eqn{b}) that may have different edge
#' densities.  The nodes of \eqn{G'} are split randomly into two halves;
#' Algorithm 2 is applied to compute an observed statistic.  Node subsampling
#' from \eqn{G} (again via Algorithm 2) generates a reference distribution.
#' A final multivariate test---either a Mahalanobis distance test or the
#' Cauchy combination test of Liu and Xie (2020)---yields a p-value.  This
#' is the first subsampling-based procedure that is valid for unmatchable
#' networks with unequal densities.
#'
#' @section Motif support:
#' Three motifs are built in: \code{"v_shape"} (2-star, 2 edges),
#' \code{"triangle"} (3 edges), and \code{"star3"} (3-star, 3 edges).
#' All three functions accept a \code{motifs} argument that can include
#' user-supplied specifications: a list with element \code{fn}, a
#' \code{function(A)} returning a \emph{named} numeric vector of one or more
#' moment values, and element \code{n_edges}, an integer vector of the
#' corresponding edge counts used for density normalisation.
#'
#' @section Parallel execution:
#' The \eqn{N_{\mathrm{sub}}} subsampling loop in \code{subsample_moments}
#' and \code{two_sample_test} can be parallelised by setting
#' \code{parallel = TRUE} (uses \pkg{doParallel} and \pkg{foreach}).
#' Parallel mode requires the \pkg{netsubsamp} package to be installed on
#' all worker nodes.
#'
#' @references
#' Qi M, Hua C-W, Li T, Zhou W (2024).
#' Multivariate Inference of Network Moments by Subsampling.
#' \emph{Biometrika}, \strong{111}(1), 1--18.
#' \href{https://arxiv.org/abs/2409.01599}{arXiv:2409.01599}.
#'
#' Liu Y, Xie J (2020).
#' Cauchy combination test: a powerful test with analytic p-value calculation
#' under arbitrary dependency structures.
#' \emph{Journal of the American Statistical Association}, \strong{115},
#' 393--402.
#'
#' @examples
#' \donttest{
#' library(randnet)
#'
#' ## ---- Algorithm 1: approximate the joint moment distribution ----------
#' set.seed(1)
#' G   <- BlockModel.Gen(lambda = 15, n = 200, K = 3)$A
#' res <- subsample_moments(G, b = 60, N_sub = 500)
#'
#' # Estimated density
#' res$rho
#'
#' # First few rows of the N_sub x 3 moment matrix
#' head(res$moments)
#'
#'
#' ## ---- Algorithm 2: externally sparsified moments ----------------------
#' set.seed(2)
#' G1  <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A
#' G2  <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A
#' sparsify_moments(G1, G2, rho_target = 0.08)
#'
#'
#' ## ---- Algorithm 3: two-sample test under unequal densities -----------
#' set.seed(3)
#' G       <- BlockModel.Gen(lambda = 20, n = 400, K = 3)$A
#' G_prime <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A
#'
#' # Choose rho_target as 0.7 * min(density(G), density(G_prime))
#' rho_G       <- sum(G)       / (400 * 399)
#' rho_G_prime <- sum(G_prime) / (100 * 99)
#' rho_target  <- 0.7 * min(rho_G, rho_G_prime)
#'
#' res <- two_sample_test(G, G_prime,
#'                        rho_target = rho_target,
#'                        N_sub      = 500,
#'                        test       = "mahalanobis")
#' res$p_value
#' }
#'
#' @docType package
#' @name netsubsamp-package
#' @aliases netsubsamp
"_PACKAGE"

# ============================================================
# Internal helper: random edge removal (sparsification).
# ============================================================


#' Randomly remove edges from an undirected network
#'
#' Each edge in the undirected network is independently retained with
#' probability \code{p} (equivalently, removed with probability
#' \code{1 - p}).  The operation generates a symmetric Bernoulli mask on the
#' upper triangle and mirrors it, preserving the symmetry of the adjacency
#' matrix.
#'
#' @param A A square symmetric binary matrix (\strong{dense}).
#' @param p Numeric in \eqn{[0, 1]}; the probability of retaining each edge.
#'
#' @return A dense integer matrix of the same dimensions as \code{A}.
#'
#' @importFrom stats rbinom
#' @keywords internal
sparsify_graph <- function(A, p) {

  if (p >= 1) return(A)
  if (p <= 0) return(matrix(0L, nrow(A), ncol(A)))

  n       <- nrow(A)
  n_upper <- n * (n - 1L) / 2L

  mask              <- matrix(0L, n, n)
  mask[upper.tri(mask)] <- stats::rbinom(n_upper, 1L, p)
  mask              <- mask + t(mask)

  A * mask
}

# Helper utilities shared across all test files.

#' Generate a random undirected binary adjacency matrix
#'
#' @param n Number of nodes.
#' @param p Edge probability.
#' @param seed Optional random seed.
make_adj <- function(n, p, seed = NULL) {
  if (!is.null(seed)) set.seed(seed)
  A <- matrix(0L, n, n)
  A[upper.tri(A)] <- rbinom(n * (n - 1L) / 2L, 1L, p)
  A + t(A)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validate_adjacency}
\alias{validate_adjacency}
\title{Validate an adjacency matrix}
\usage{
validate_adjacency(A, name = "A")
}
\arguments{
\item{A}{Object to validate.}

\item{name}{Name used in error messages.}
}
\description{
Checks that \code{A} is a square matrix (dense or sparse \code{Matrix})
and warns if it does not appear symmetric.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/two_sample_test.R
\name{two_sample_test}
\alias{two_sample_test}
\title{Two-sample test for unmatchable networks with unequal densities (Algorithm 3)}
\usage{
two_sample_test(
  G,
  G_prime,
  rho_target,
  N_sub = 2000L,
  motifs = list("v_shape", "triangle", "star3"),
  test = c("mahalanobis", "cauchy"),
  parallel = FALSE,
  n_cores = NULL
)
}
\arguments{
\item{G}{A square symmetric binary adjacency matrix (dense or sparse
\code{Matrix}) of size \eqn{n \times n}.  Typically the larger or
reference network.}

\item{G_prime}{A square symmetric binary adjacency matrix of size
\eqn{b \times b}.  The network to compare against \code{G}.}

\item{rho_target}{Numeric in \eqn{(0, 1)}; the common target edge density
\eqn{\rho^\dagger}.  See the \emph{Choice of \code{rho_target}} section.}

\item{N_sub}{Integer; number of subsampling replicates.  Default 2000.}

\item{motifs}{A list of motif specifications; see
\code{\link{subsample_moments}} for the format.  Defaults to
\code{list("v_shape", "triangle", "star3")}.}

\item{test}{Character; the multivariate test to apply.  Either
\code{"mahalanobis"} (default) or \code{"cauchy"}.}

\item{parallel}{Logical; whether to parallelise the subsampling loop.
Default \code{FALSE}.  Requires the \pkg{netsubsamp} package to be
\emph{installed} on all worker nodes when \code{TRUE}.}

\item{n_cores}{Integer or \code{NULL}; number of cores when
\code{parallel = TRUE}.}
}
\value{
A list with:
\describe{
  \item{\code{p_value}}{The p-value of the test.}
  \item{\code{test_statistic}}{Observed test statistic: Mahalanobis
    distance \eqn{D_0} when \code{test = "mahalanobis"}, or Cauchy
    combination statistic \eqn{T} when \code{test = "cauchy"}.}
  \item{\code{observed_moments}}{Named numeric vector; the observed
    \eqn{\bar\Psi} statistic.}
  \item{\code{null_moments}}{\eqn{N_{\mathrm{sub}} \times m} matrix of
    null \eqn{\bar\Psi} statistics (including any \code{NA} rows from
    degenerate subsamples).}
  \item{\code{null_statistic}}{Numeric vector of per-replicate null
    Mahalanobis distances (or marginal p-values for \code{"cauchy"}).}
  \item{\code{marginal_p_values}}{(Only for \code{test = "cauchy"}.) Named
    numeric vector of per-motif marginal p-values.}
  \item{\code{method}}{The test method used.}
  \item{\code{motif_names}}{Character vector of moment names.}
}
}
\description{
Tests whether two unmatchable networks --- differing in node set, size, and
possibly edge density --- arise from the same underlying graphon model.
The procedure avoids the density-mismatch problem via a
\emph{split-and-sparsify} strategy: each network is randomly split into two
halves, one used to estimate the sparsification probability and the other to
compute density-normalised moments at a shared target density.  Node
subsampling from the larger network then generates a reference distribution
against which the observed statistic is compared.
}
\section{Algorithm}{

\enumerate{
  \item Randomly partition the nodes of \code{G_prime} into two equal
    halves \eqn{G'_1} and \eqn{G'_2}; compute the observed statistic
    \eqn{\bar\Psi_{\rho^\dagger}(G'_1, G'_2)} via Algorithm 2.
  \item For \eqn{i = 1, \ldots, N_{\mathrm{sub}}}: independently draw
    two sets of \eqn{\lfloor b/2 \rfloor} nodes from \eqn{G} and compute
    \eqn{\bar\Psi_{\rho^\dagger}(G^*_{i1}, G^*_{i2})} via Algorithm 2.
  \item Compare the observed statistic against the empirical null
    distribution using the selected \code{test}.
}
}

\section{Choice of \code{rho_target}}{

A recommended default is
\eqn{\rho^\dagger = \kappa \cdot \min(\hat\rho_G, \hat\rho_{G'})},
with \eqn{\kappa \in (0.5,\, 0.9)}.  Values close to 0 discard too many
edges and inflate variance; values too close to
\eqn{\min(\hat\rho_G, \hat\rho_{G'})} leave little room for edge removal
and can break the density-matching property.  A value of \eqn{\kappa = 0.7}
(corresponding to
\code{rho_target = 0.7 * min(rho_hat_G, rho_hat_G_prime)}) works well
across a range of settings.
}

\section{Tests}{

\describe{
  \item{Mahalanobis (default)}{Computes a multivariate distance that fully
    exploits the joint distribution of all moments.  The p-value is the
    proportion of null distances exceeding the observed distance.  This test
    is more powerful when the signal lies in the \emph{joint} rather than
    the \emph{marginal} distribution of the moments.}
  \item{Cauchy combination}{Combines marginal p-values via the Cauchy
    combination test of Liu & Xie (2020).  Controls type I error under
    arbitrary dependence but ignores cross-moment correlation.}
}
}

\examples{
\donttest{
library(randnet)
set.seed(3)
G       <- BlockModel.Gen(lambda = 20, n = 400, K = 3)$A
G_prime <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A

# Choose rho_target as 0.7 * min(density(G), density(G_prime))
rho_G       <- sum(G)       / (400 * 399)
rho_G_prime <- sum(G_prime) / (100 * 99)
rho_target  <- 0.7 * min(rho_G, rho_G_prime)

res <- two_sample_test(G, G_prime,
                       rho_target = rho_target,
                       N_sub      = 500,
                       test       = "mahalanobis")
res$p_value
}

}
\references{
Qi, Hua, Li and Zhou (2024). Multivariate Inference of Network Moments by
Subsampling. \emph{Biometrika}, \strong{111}(1), 1--18.
\doi{10.1093/biomet/asad056}.

Liu, Y. and Xie, J. (2020). Cauchy combination test: a powerful test with
analytic p-value calculation under arbitrary dependency structures.
\emph{Journal of the American Statistical Association}, \strong{115},
393--402. \doi{10.1080/01621459.2018.1554485}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_moments.R
\name{compute_network_moments}
\alias{compute_network_moments}
\title{Compute edge density and network moments}
\usage{
compute_network_moments(A, motif_specs)
}
\arguments{
\item{A}{A square binary adjacency matrix (\strong{dense}).  Must have at
least 3 rows/columns for V-shape and triangle moments, and at least 4 for
the 3-star moment.}

\item{motif_specs}{Compiled motif specification, as returned by
\code{\link{init_motif_specs}}.}
}
\value{
A list with:
\describe{
  \item{\code{rho}}{Scalar edge density.}
  \item{\code{UR}}{Named numeric vector of network moments (one per motif
    in \code{motif_specs}).}
}
Rows corresponding to graphs with zero density return \code{rho = 0} and
all moments equal to zero.
}
\description{
Given a square binary adjacency matrix and a compiled motif specification,
returns the edge density and the vector of network moments (rescaled motif
counts).
}
\details{
For a network with \eqn{k} nodes the edge density is
\eqn{\hat\rho = |E| / [k(k-1)]}.  The network moment for motif \eqn{R}
with \eqn{r} nodes is \eqn{U_R = \binom{k}{r}^{-1} X_R}, where \eqn{X_R}
is the (non-induced) motif count.  Built-in moments use matrix-algebraic
formulas that avoid explicit enumeration.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{setup_foreach}
\alias{setup_foreach}
\title{Set up a foreach parallel backend}
\usage{
setup_foreach(parallel, n_cores)
}
\arguments{
\item{parallel}{Logical; whether parallelism is requested.}

\item{n_cores}{Integer or \code{NULL}.  When \code{parallel = TRUE} and
\code{n_cores} is \code{NULL}, one fewer than the number of detected
cores is used (minimum 1).}
}
\value{
A list with:
\describe{
  \item{\code{op}}{The foreach binary operator (\code{\%do\%} or
    \code{\%dopar\%}).}
  \item{\code{cleanup}}{A zero-argument function that stops any cluster
    created here and re-registers the sequential backend.}
}
}
\description{
Registers a \pkg{doParallel} backend when requested and returns the
appropriate foreach binary operator together with a cleanup function.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparsify_graph.R
\name{sparsify_graph}
\alias{sparsify_graph}
\title{Randomly remove edges from an undirected network}
\usage{
sparsify_graph(A, p)
}
\arguments{
\item{A}{A square symmetric binary matrix (\strong{dense}).}

\item{p}{Numeric in \eqn{[0, 1]}; the probability of retaining each edge.}
}
\value{
A dense integer matrix of the same dimensions as \code{A}.
}
\description{
Each edge in the undirected network is independently retained with
probability \code{p} (equivalently, removed with probability
\code{1 - p}).  The operation generates a symmetric Bernoulli mask on the
upper triangle and mirrors it, preserving the symmetry of the adjacency
matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subsample_moments.R
\name{subsample_moments}
\alias{subsample_moments}
\title{Uniform node subsampling for multivariate network moments (Algorithm 1)}
\usage{
subsample_moments(
  G,
  b,
  N_sub,
  motifs = list("v_shape", "triangle", "star3"),
  parallel = FALSE,
  n_cores = NULL
)
}
\arguments{
\item{G}{A square symmetric binary adjacency matrix (dense \code{matrix}
or sparse \code{Matrix}) of size \eqn{n \times n}.}

\item{b}{Integer; the subsampling size.  Must satisfy \eqn{3 \le b < n}.}

\item{N_sub}{Integer; number of subsamples to draw.}

\item{motifs}{A list of motif specifications.  Each element is either:
\itemize{
  \item a character string: \code{"v_shape"}, \code{"triangle"}, or
    \code{"star3"} (built-in defaults); or
  \item a named list with elements \code{fn} -- a
    \code{function(A)} returning a \emph{named} numeric vector -- and
    \code{n_edges} -- an integer vector of edge counts matching the
    output length of \code{fn}.
}
Defaults to \code{list("v_shape", "triangle", "star3")}.}

\item{parallel}{Logical; if \code{TRUE} the subsampling loop is run in
parallel via \pkg{doParallel} and \pkg{foreach}.  Default \code{FALSE}.
Requires the \pkg{netsubsamp} package to be \emph{installed} on all
worker nodes when \code{TRUE}.}

\item{n_cores}{Integer or \code{NULL}; number of cores to use when
\code{parallel = TRUE}.  If \code{NULL}, uses
\code{parallel::detectCores() - 1} (minimum 1).}
}
\value{
A list with:
\describe{
  \item{\code{rho}}{Estimated edge density of \code{G}.}
  \item{\code{moments}}{An \eqn{N_{\mathrm{sub}} \times m} numeric matrix
    of subsampled network moments.  Rows for which the subsampled graph has
    zero density contain \code{NA}.}
  \item{\code{motif_names}}{Character vector of moment names, in column
    order of \code{moments}.}
  \item{\code{b}}{The subsampling size used.}
}
}
\description{
Repeatedly draws random node-induced subgraphs of size \code{b} from
network \code{G} and computes network moments for each subsample.  The
resulting empirical distribution approximates the joint distribution of
network moments for networks of size \code{b} drawn from the same
underlying graphon model, providing the foundation for downstream inference.
}
\examples{
\donttest{
library(randnet)
set.seed(1)
G   <- BlockModel.Gen(lambda = 15, n = 200, K = 3)$A
res <- subsample_moments(G, b = 60, N_sub = 500)
head(res$moments)
res$rho
}

}
\references{
Qi, Hua, Li and Zhou (2024). Multivariate Inference of Network Moments by
Subsampling. \emph{Biometrika}, \strong{111}(1), 1--18.
\doi{10.1093/biomet/asad056}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparsify_moments.R
\name{.sparsify_moments_internal}
\alias{.sparsify_moments_internal}
\title{Compute externally sparsified network moments (internal)}
\usage{
.sparsify_moments_internal(G1, G2, rho_target, motif_specs)
}
\arguments{
\item{G1}{Dense binary adjacency matrix used as the density estimator.}

\item{G2}{Dense binary adjacency matrix used as the moment source.}

\item{rho_target}{Numeric; target edge density \eqn{\rho^\dagger \in (0,1)}.}

\item{motif_specs}{Compiled motif specification from
\code{\link{init_motif_specs}}.}
}
\value{
Named numeric vector of density-normalised moments
  \eqn{\hat\rho_{\tilde{G}_2}^{-|E(R)|} \cdot U_R(\tilde{G}_2)}, or a
  vector of \code{NA} values if the result is degenerate (zero density
  after edge removal).
}
\description{
Uses \code{G1} to estimate the sparsification probability and
independently applies random edge removal to \code{G2} to reach the
target density \code{rho_target}.  Returns the density-normalised moment
vector \eqn{\bar\Psi_{\rho^\dagger}(G_1, G_2)}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{init_motif_specs}
\alias{init_motif_specs}
\title{Initialise and validate a list of motif specifications}
\usage{
init_motif_specs(motifs)
}
\arguments{
\item{motifs}{A list of motif specifications (see above).}
}
\value{
A list with three elements:
\describe{
  \item{\code{compute}}{A function \code{function(A)} that evaluates all
    motifs on an adjacency matrix and returns a named numeric vector.}
  \item{\code{edge_counts}}{Integer vector of edge counts, one per moment.}
  \item{\code{names}}{Character vector of moment names, in the same order
    as \code{edge_counts}.}
}
}
\description{
Parses a mixed list of built-in motif names (character strings) and
user-supplied specifications (lists with \code{fn} and \code{n_edges})
into an internal representation used throughout the package.
}
\details{
Built-in motifs and their edge counts:
\itemize{
  \item \code{"v_shape"} -- 2-star (V-shape), 2 edges
  \item \code{"triangle"} -- triangle, 3 edges
  \item \code{"star3"} -- 3-star, 3 edges
}

A user-supplied specification must be a named list with:
\itemize{
  \item \code{fn}: a \code{function(A)} that accepts a square binary
    adjacency matrix and returns a \emph{named} numeric vector of one or
    more moment values (one element per motif computed by that function).
  \item \code{n_edges}: integer vector whose length equals the length of
    \code{fn}'s return value, giving the number of edges in each
    corresponding motif.  Used for density normalisation.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/netsubsamp-package.R
\docType{package}
\name{netsubsamp-package}
\alias{netsubsamp-package}
\alias{netsubsamp}
\title{netsubsamp: Multivariate Inference of Network Moments by Subsampling}
\description{
\pkg{netsubsamp} implements the three subsampling algorithms from
Qi, Hua, Li and Zhou (2024) for multivariate inference on network moments
(rescaled counts of graph motifs such as V-shapes, triangles, and stars)
under a sparse graphon model.
}
\section{The three algorithms}{


\strong{Algorithm 1 -- \code{\link{subsample_moments}}: Uniform node
subsampling.}
Draws \eqn{N_{\mathrm{sub}}} random node-induced subgraphs of size
\eqn{b} from a single network \eqn{G} and computes the vector of network
moments for each subsample.  The resulting empirical distribution
consistently approximates the joint distribution of multiple network moments
for networks of size \eqn{b} drawn from the same graphon.  This provides
the foundation for any downstream inference task: confidence regions,
goodness-of-fit tests, or visualization of the joint moment distribution.

\strong{Algorithm 2 -- \code{\link{sparsify_moments}}: Externally
sparsified moments.}
Given two independently sampled subgraphs \eqn{G_1} and \eqn{G_2} and a
target density \eqn{\rho^\dagger}, uses \eqn{G_1} to estimate the
required edge-removal probability and then randomly removes edges from
\eqn{G_2} to produce a graph at the target density.  Returns the
density-normalised moment vector
\eqn{\bar\Psi_{\rho^\dagger}(G_1, G_2)}.  This building block handles
the practical complication that the true graphon density is unknown.

\strong{Algorithm 3 -- \code{\link{two_sample_test}}: Two-sample test for
unmatchable networks with unequal densities.}
Tests \eqn{H_0: w = w'} for two unmatchable networks \eqn{G} (size
\eqn{n}) and \eqn{G'} (size \eqn{b}) that may have different edge
densities.  The nodes of \eqn{G'} are split randomly into two halves;
Algorithm 2 is applied to compute an observed statistic.  Node subsampling
from \eqn{G} (again via Algorithm 2) generates a reference distribution.
A final multivariate test---either a Mahalanobis distance test or the
Cauchy combination test of Liu and Xie (2020)---yields a p-value.  This
is the first subsampling-based procedure that is valid for unmatchable
networks with unequal densities.
}

\section{Motif support}{

Three motifs are built in: \code{"v_shape"} (2-star, 2 edges),
\code{"triangle"} (3 edges), and \code{"star3"} (3-star, 3 edges).
All three functions accept a \code{motifs} argument that can include
user-supplied specifications: a list with element \code{fn}, a
\code{function(A)} returning a \emph{named} numeric vector of one or more
moment values, and element \code{n_edges}, an integer vector of the
corresponding edge counts used for density normalisation.
}

\section{Parallel execution}{

The \eqn{N_{\mathrm{sub}}} subsampling loop in \code{subsample_moments}
and \code{two_sample_test} can be parallelised by setting
\code{parallel = TRUE} (uses \pkg{doParallel} and \pkg{foreach}).
Parallel mode requires the \pkg{netsubsamp} package to be installed on
all worker nodes.
}

\examples{
\donttest{
library(randnet)

## ---- Algorithm 1: approximate the joint moment distribution ----------
set.seed(1)
G   <- BlockModel.Gen(lambda = 15, n = 200, K = 3)$A
res <- subsample_moments(G, b = 60, N_sub = 500)

# Estimated density
res$rho

# First few rows of the N_sub x 3 moment matrix
head(res$moments)


## ---- Algorithm 2: externally sparsified moments ----------------------
set.seed(2)
G1  <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A
G2  <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A
sparsify_moments(G1, G2, rho_target = 0.08)


## ---- Algorithm 3: two-sample test under unequal densities -----------
set.seed(3)
G       <- BlockModel.Gen(lambda = 20, n = 400, K = 3)$A
G_prime <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A

# Choose rho_target as 0.7 * min(density(G), density(G_prime))
rho_G       <- sum(G)       / (400 * 399)
rho_G_prime <- sum(G_prime) / (100 * 99)
rho_target  <- 0.7 * min(rho_G, rho_G_prime)

res <- two_sample_test(G, G_prime,
                       rho_target = rho_target,
                       N_sub      = 500,
                       test       = "mahalanobis")
res$p_value
}

}
\references{
Qi M, Hua C-W, Li T, Zhou W (2024).
Multivariate Inference of Network Moments by Subsampling.
\emph{Biometrika}, \strong{111}(1), 1--18.
\href{https://arxiv.org/abs/2409.01599}{arXiv:2409.01599}.

Liu Y, Xie J (2020).
Cauchy combination test: a powerful test with analytic p-value calculation
under arbitrary dependency structures.
\emph{Journal of the American Statistical Association}, \strong{115},
393--402.
}
\author{
\strong{Maintainer}: Tianxi Li \email{tianxili@umn.edu}

Authors:
\itemize{
  \item Mingyu Qi
  \item Chen-Wei Hua
  \item Wen Zhou
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparsify_moments.R
\name{sparsify_moments}
\alias{sparsify_moments}
\title{Externally sparsified network moments (Algorithm 2)}
\usage{
sparsify_moments(
  G1,
  G2,
  rho_target,
  motifs = list("v_shape", "triangle", "star3")
)
}
\arguments{
\item{G1}{A square symmetric binary adjacency matrix (dense or sparse
\code{Matrix}) used to estimate the sparsification probability.}

\item{G2}{A square symmetric binary adjacency matrix from which moments
are computed after edge removal.}

\item{rho_target}{Numeric in \eqn{(0, 1)}; the target edge density
\eqn{\rho^\dagger}.}

\item{motifs}{A list of motif specifications; see
\code{\link{subsample_moments}} for the format.  Defaults to
\code{list("v_shape", "triangle", "star3")}.}
}
\value{
A named numeric vector of density-normalised network moments, or
  \code{NA} values if the graph after edge removal has zero density.
}
\description{
Uses one adjacency matrix (\code{G1}) to estimate the sparsification
probability and a second independent adjacency matrix (\code{G2}) to
compute density-normalised network moments after random edge removal to a
common target density.  The resulting statistic
\eqn{\bar\Psi_{\rho^\dagger}(G_1, G_2)} is the building block for the
two-sample test in \code{\link{two_sample_test}}.
}
\section{Algorithm}{

\enumerate{
  \item Estimate \eqn{\hat{p} = \min\!\left(1,\,\rho^\dagger /
    \hat\rho_{G_1}\right)}.
  \item Randomly remove edges from \eqn{G_2}: each edge is retained
    independently with probability \eqn{\hat{p}}, yielding
    \eqn{\tilde{G}_2}.
  \item Return \eqn{\hat\rho_{\tilde{G}_2}^{-|E(R)|} \cdot
    U_R(\tilde{G}_2)} for each motif \eqn{R}.
}
}

\examples{
\donttest{
library(randnet)
set.seed(2)
G1 <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A
G2 <- BlockModel.Gen(lambda = 15, n = 100, K = 3)$A
sparsify_moments(G1, G2, rho_target = 0.08)
}

}
\references{
Qi, Hua, Li and Zhou (2024). Multivariate Inference of Network Moments by
Subsampling. \emph{Biometrika}, \strong{111}(1), 1--18.
\doi{10.1093/biomet/asad056}.
}
